#         Based on the original rnsh program by Aaron Heise (@acehoss)
# https://github.com/acehoss/rnsh - MIT License - Copyright (c) 2023 Aaron Heise
#     This version of rnsh is included in RNS under the Reticulum License
#
# Reticulum License
#
# Copyright (c) 2016-2026 Mark Qvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# - The Software shall not be used in any kind of system which includes amongst
#   its functions the ability to purposefully do harm to human beings.
#
# - The Software shall not be used, directly or indirectly, in the creation of
#   an artificial intelligence, machine learning or language model training
#   dataset, including but not limited to any use that contributes to the
#   training or development of such a model or algorithm.
#
# - The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import argparse
import sys

from RNS.Utilities.rnsh._version import __version__ as __rnsh_version__
from RNS._version import __version__

DEFAULT_SERVICE_NAME = "default"

def setup_argument_parser():
    parser = argparse.ArgumentParser(description="Reticulum Remote Shell Utility", epilog="When specifying a command to execute, separate rnsh\noptions from the command and its arguments with --\n\nFor example:\n  rnsh -l -- /bin/bash --login\n  rnsh <destination> -- ls -la /tmp", formatter_class=argparse.RawDescriptionHelpFormatter)

    # Common options
    parser.add_argument("--config", "-c", action="store", default=None, help="path to alternative Reticulum config directory", type=str)
    parser.add_argument("--identity", "-i", action="store", default=None, help="path to identity file to use", type=str)
    parser.add_argument("-v", "--verbose", action="count", default=0, help="increase verbosity")
    parser.add_argument("-q", "--quiet", action="count", default=0, help="decrease verbosity")
    parser.add_argument("-p", "--print-identity", action="store_true", default=False, help="print identity and destination info and exit")
    parser.add_argument("--version", action="version", version="rnsh {rv} (protocol {pv})".format(rv=__version__, pv=__rnsh_version__))

    # Listener options
    parser.add_argument("-l", "--listen", action="store_true", default=False, help="listen (server) mode; any command specified after -- will be used as the default command when the initiator does not provide one or when remote command execution is disabled; if no command is specified, the default shell of the user running rnsh will be used")
    parser.add_argument("-s", "--service", action="store", default=None, help="service name for identity file if not the default", type=str)
    parser.add_argument("-b", "--announce",action="store", default=None,help="announce on startup and every PERIOD seconds; specify 0 to announce on startup only",metavar="PERIOD", type=int)
    parser.add_argument("-a", "--allowed", action="append", default=None, metavar="HASH", type=str, help="allow this identity to connect (may be specified multiple times); allowed identities can also be specified in ~/.rnsh/allowed_identities or ~/.config/rnsh/allowed_identities, one hash per line")
    parser.add_argument("-n", "--no-auth", action="store_true", default=False, help="disable authentication (allow any identity to connect)")
    parser.add_argument("-A", "--remote-command-as-args", action="store_true", default=False, help="concatenate remote command to the argument list of the default program or shell")
    parser.add_argument("-C", "--no-remote-command", action="store_true", default=False, help="disable executing command lines received from the remote initiator")

    # Initiator options
    parser.add_argument("-N", "--no-id", action="store_true", default=False, help="disable identity announcement on connect")
    parser.add_argument("-m", "--mirror", action="store_true", default=False, help="return with the exit code of the remote process")
    parser.add_argument("-w", "--timeout", action="store", default=None, help="connect and request timeout in seconds", metavar="SECONDS", type=float)

    parser.add_argument("destination", nargs="?", default=None, help="hexadecimal hash of the destination to connect to", type=str)

    return parser


def parse_arguments(argv=None):
    if argv is None: argv = sys.argv[1:]

    # Split at -- to separate rnsh options from the command to execute.
    # Everything before -- (or the entire argv if no --) goes to argparse.
    # Everything after -- becomes the command list.
    try:
        split_idx = argv.index("--")
        rnsh_argv = argv[:split_idx]
        command = argv[split_idx + 1:]
    except ValueError:
        rnsh_argv = argv
        command = []

    parser = setup_argument_parser()
    args = parser.parse_args(rnsh_argv)
    args.command = command

    if args.listen and not args.service: args.service = DEFAULT_SERVICE_NAME

    return args, parser



# Reticulum License
#
# Copyright (c) 2016-2025 Mark Qvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# - The Software shall not be used in any kind of system which includes amongst
#   its functions the ability to purposefully do harm to human beings.
#
# - The Software shall not be used, directly or indirectly, in the creation of
#   an artificial intelligence, machine learning or language model training
#   dataset, including but not limited to any use that contributes to the
#   training or development of such a model or algorithm.
#
# - The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from RNS.Interfaces.Interface import Interface
from time import sleep
import sys
import threading
import time
import RNS

class HDLC():
    # The Serial Interface packetizes data using
    # simplified HDLC framing, similar to PPP
    FLAG              = 0x7E
    ESC               = 0x7D
    ESC_MASK          = 0x20

    @staticmethod
    def escape(data):
        data = data.replace(bytes([HDLC.ESC]), bytes([HDLC.ESC, HDLC.ESC^HDLC.ESC_MASK]))
        data = data.replace(bytes([HDLC.FLAG]), bytes([HDLC.ESC, HDLC.FLAG^HDLC.ESC_MASK]))
        return data

class SerialInterface(Interface):
    MAX_CHUNK = 32768
    DEFAULT_IFAC_SIZE = 8

    owner    = None
    port     = None
    speed    = None
    databits = None
    parity   = None
    stopbits = None
    serial   = None

    def __init__(self, owner, configuration):
        import importlib.util
        if RNS.vendor.platformutils.is_android():
            self.on_android  = True
            if importlib.util.find_spec('usbserial4a') != None:
                if importlib.util.find_spec('jnius') == None:
                    RNS.log("Could not load jnius API wrapper for Android, Serial interface cannot be created.", RNS.LOG_CRITICAL)
                    RNS.log("This probably means you are trying to use an USB-based interface from within Termux or similar.", RNS.LOG_CRITICAL)
                    RNS.log("This is currently not possible, due to this environment limiting access to the native Android APIs.", RNS.LOG_CRITICAL)
                    RNS.panic()

                from usbserial4a import serial4a as serial
                self.parity = "N"
            
            else:
                RNS.log("Could not load USB serial module for Android, Serial interface cannot be created.", RNS.LOG_CRITICAL)
                RNS.log("You can install this module by issuing: pip install usbserial4a", RNS.LOG_CRITICAL)
                RNS.panic()
        else:
            raise SystemError("Android-specific interface was used on non-Android OS")

        super().__init__()

        c = Interface.get_config_obj(configuration)
        name = c["name"]
        port = c["port"] if "port" in c else None
        speed = int(c["speed"]) if "speed" in c else 9600
        databits = int(c["databits"]) if "databits" in c else 8
        parity = c["parity"] if "parity" in c else "N"
        stopbits = int(c["stopbits"]) if "stopbits" in c else 1

        if port == None:
            raise ValueError("No port specified for serial interface")

        self.HW_MTU = 564
        
        self.pyserial = serial
        self.serial   = None
        self.owner    = owner
        self.name     = name
        self.port     = port
        self.speed    = speed
        self.databits = databits
        self.parity   = "N"
        self.stopbits = stopbits
        self.timeout  = 100
        self.online   = False
        self.bitrate  = self.speed

        if parity.lower() == "e" or parity.lower() == "even":
            self.parity = "E"

        if parity.lower() == "o" or parity.lower() == "odd":
            self.parity = "O"

        try:
            self.open_port()
        except Exception as e:
            RNS.log("Could not open serial port for interface "+str(self), RNS.LOG_ERROR)
            raise e

        if self.serial.is_open:
            self.configure_device()
        else:
            raise IOError("Could not open serial port")


    def open_port(self):
        RNS.log("Opening serial port "+self.port+"...")
        # Get device parameters
        from usb4a import usb
        device = usb.get_usb_device(self.port)
        if device:
            vid = device.getVendorId()
            pid = device.getProductId()

            # Driver overrides for speficic chips
            proxy = self.pyserial.get_serial_port
            if vid == 0x1A86 and pid == 0x55D4:
                # Force CDC driver for Qinheng CH34x
                RNS.log(str(self)+" using CDC driver for "+RNS.hexrep(vid)+":"+RNS.hexrep(pid), RNS.LOG_DEBUG)
                from usbserial4a.cdcacmserial4a import CdcAcmSerial
                proxy = CdcAcmSerial

            self.serial = proxy(
                self.port,
                baudrate = self.speed,
                bytesize = self.databits,
                parity = self.parity,
                stopbits = self.stopbits,
                xonxoff = False,
                rtscts = False,
                timeout = None,
                inter_byte_timeout = None,
                # write_timeout = wtimeout,
                dsrdtr = False,
            )

            if vid == 0x0403:
                # Hardware parameters for FTDI devices @ 115200 baud
                self.serial.DEFAULT_READ_BUFFER_SIZE = 16 * 1024
                self.serial.USB_READ_TIMEOUT_MILLIS = 100
                self.serial.timeout = 0.1
            elif vid == 0x10C4:
                # Hardware parameters for SiLabs CP210x @ 115200 baud
                self.serial.DEFAULT_READ_BUFFER_SIZE = 64 
                self.serial.USB_READ_TIMEOUT_MILLIS = 12
                self.serial.timeout = 0.012
            elif vid == 0x1A86 and pid == 0x55D4:
                # Hardware parameters for Qinheng CH34x @ 115200 baud
                self.serial.DEFAULT_READ_BUFFER_SIZE = 64
                self.serial.USB_READ_TIMEOUT_MILLIS = 12
                self.serial.timeout = 0.1
            else:
                # Default values
                self.serial.DEFAULT_READ_BUFFER_SIZE = 1 * 1024
                self.serial.USB_READ_TIMEOUT_MILLIS = 100
                self.serial.timeout = 0.1

            RNS.log(str(self)+" USB read buffer size set to "+RNS.prettysize(self.serial.DEFAULT_READ_BUFFER_SIZE), RNS.LOG_DEBUG)
            RNS.log(str(self)+" USB read timeout set to "+str(self.serial.USB_READ_TIMEOUT_MILLIS)+"ms", RNS.LOG_DEBUG)
            RNS.log(str(self)+" USB write timeout set to "+str(self.serial.USB_WRITE_TIMEOUT_MILLIS)+"ms", RNS.LOG_DEBUG)

    def configure_device(self):
        sleep(0.5)
        thread = threading.Thread(target=self.readLoop)
        thread.daemon = True
        thread.start()
        self.online = True
        RNS.log("Serial port "+self.port+" is now open", RNS.LOG_VERBOSE)


    def process_incoming(self, data):
        self.rxb += len(data)
        def af():
            self.owner.inbound(data, self)
        threading.Thread(target=af, daemon=True).start()

    def process_outgoing(self,data):
        if self.online:
            data = bytes([HDLC.FLAG])+HDLC.escape(data)+bytes([HDLC.FLAG])
            written = self.serial.write(data)
            self.txb += len(data)            
            if written != len(data):
                raise IOError("Serial interface only wrote "+str(written)+" bytes of "+str(len(data)))

    def readLoop(self):
        try:
            in_frame = False
            escape = False
            data_buffer = b""
            last_read_ms = int(time.time()*1000)

            while self.serial.is_open:
                serial_bytes = self.serial.read()
                got = len(serial_bytes)

                for byte in serial_bytes:
                    last_read_ms = int(time.time()*1000)

                    if (in_frame and byte == HDLC.FLAG):
                        in_frame = False
                        self.process_incoming(data_buffer)
                    elif (byte == HDLC.FLAG):
                        in_frame = True
                        data_buffer = b""
                    elif (in_frame and len(data_buffer) < self.HW_MTU):
                        if (byte == HDLC.ESC):
                            escape = True
                        else:
                            if (escape):
                                if (byte == HDLC.FLAG ^ HDLC.ESC_MASK):
                                    byte = HDLC.FLAG
                                if (byte == HDLC.ESC  ^ HDLC.ESC_MASK):
                                    byte = HDLC.ESC
                                escape = False
                            data_buffer = data_buffer+bytes([byte])
                        
                if got == 0:
                    time_since_last = int(time.time()*1000) - last_read_ms
                    if len(data_buffer) > 0 and time_since_last > self.timeout:
                        data_buffer = b""
                        in_frame = False
                        escape = False
                    # sleep(0.08)
                    
        except Exception as e:
            self.online = False
            RNS.log("A serial port error occurred, the contained exception was: "+str(e), RNS.LOG_ERROR)
            RNS.log("The interface "+str(self)+" experienced an unrecoverable error and is now offline.", RNS.LOG_ERROR)
            
            if RNS.Reticulum.panic_on_interface_error:
                RNS.panic()

            RNS.log("Reticulum will attempt to reconnect the interface periodically.", RNS.LOG_ERROR)

        self.online = False
        self.serial.close()
        self.reconnect_port()

    def reconnect_port(self):
        while not self.online:
            try:
                time.sleep(5)
                RNS.log("Attempting to reconnect serial port "+str(self.port)+" for "+str(self)+"...", RNS.LOG_VERBOSE)
                self.open_port()
                if self.serial.is_open:
                    self.configure_device()
            except Exception as e:
                RNS.log("Error while reconnecting port, the contained exception was: "+str(e), RNS.LOG_ERROR)

        RNS.log("Reconnected serial port for "+str(self))

    def should_ingress_limit(self):
        return False

    def __str__(self):
        return "SerialInterface["+self.name+"]"



# Reticulum License
#
# Copyright (c) 2016-2025 Mark Qvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# - The Software shall not be used in any kind of system which includes amongst
#   its functions the ability to purposefully do harm to human beings.
#
# - The Software shall not be used, directly or indirectly, in the creation of
#   an artificial intelligence, machine learning or language model training
#   dataset, including but not limited to any use that contributes to the
#   training or development of such a model or algorithm.
#
# - The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from RNS.Interfaces.Interface import Interface
import socketserver
import threading
import platform
import socket
import time
import sys
import os
import RNS

class TCPInterface():
    HW_MTU            = 262144

class HDLC():
    FLAG              = 0x7E
    ESC               = 0x7D
    ESC_MASK          = 0x20

    @staticmethod
    def escape(data):
        data = data.replace(bytes([HDLC.ESC]), bytes([HDLC.ESC, HDLC.ESC^HDLC.ESC_MASK]))
        data = data.replace(bytes([HDLC.FLAG]), bytes([HDLC.ESC, HDLC.FLAG^HDLC.ESC_MASK]))
        return data

class KISS():
    FEND              = 0xC0
    FESC              = 0xDB
    TFEND             = 0xDC
    TFESC             = 0xDD
    CMD_DATA          = 0x00
    CMD_UNKNOWN       = 0xFE

    @staticmethod
    def escape(data):
        data = data.replace(bytes([0xdb]), bytes([0xdb, 0xdd]))
        data = data.replace(bytes([0xc0]), bytes([0xdb, 0xdc]))
        return data

class ThreadingTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    pass

class ThreadingTCP6Server(socketserver.ThreadingMixIn, socketserver.TCPServer):
    address_family = socket.AF_INET6

class TCPClientInterface(Interface):
    BITRATE_GUESS = 10*1000*1000
    DEFAULT_IFAC_SIZE = 16
    AUTOCONFIGURE_MTU = True

    RECONNECT_WAIT = 5
    RECONNECT_MAX_TRIES = None

    # TCP socket options
    TCP_USER_TIMEOUT = 24
    TCP_PROBE_AFTER = 5
    TCP_PROBE_INTERVAL = 2
    TCP_PROBES = 12

    INITIAL_CONNECT_TIMEOUT = 5
    SYNCHRONOUS_START = True

    I2P_USER_TIMEOUT = 45
    I2P_PROBE_AFTER = 10
    I2P_PROBE_INTERVAL = 9
    I2P_PROBES = 5

    def __init__(self, owner, configuration, connected_socket=None):
        super().__init__()

        c = Interface.get_config_obj(configuration)
        name = c["name"]
        target_ip = c["target_host"] if "target_host" in c and c["target_host"] != None else None
        target_port = int(c["target_port"]) if "target_port" in c and c["target_host"] != None else None
        kiss_framing = False
        if "kiss_framing" in c and c.as_bool("kiss_framing") == True:
            kiss_framing = True
        i2p_tunneled = c.as_bool("i2p_tunneled") if "i2p_tunneled" in c else False
        connect_timeout = c.as_int("connect_timeout") if "connect_timeout" in c else None
        max_reconnect_tries = c.as_int("max_reconnect_tries") if "max_reconnect_tries" in c else None
        fixed_mtu = c.as_int("fixed_mtu") if "fixed_mtu" in c else None
        if fixed_mtu:
            if fixed_mtu < RNS.Reticulum.MTU: raise ValueError(f"Configured MTU of {fixed_mtu} bytes is too small")
            self.AUTOCONFIGURE_MTU = False
            self.FIXED_MTU = True
        
        self.HW_MTU           = TCPInterface.HW_MTU if not fixed_mtu else fixed_mtu
        self.IN               = True
        self.OUT              = False
        self.socket           = None
        self.parent_interface = None
        self.name             = name
        self.initiator        = False
        self.reconnecting     = False
        self.never_connected  = True
        self.owner            = owner
        self.writing          = False
        self.online           = False
        self.detached         = False
        self.kiss_framing     = kiss_framing
        self.i2p_tunneled     = i2p_tunneled
        self.mode             = RNS.Interfaces.Interface.Interface.MODE_FULL
        self.bitrate          = TCPClientInterface.BITRATE_GUESS
        
        self.supports_discovery = True
        if max_reconnect_tries == None: self.max_reconnect_tries = TCPClientInterface.RECONNECT_MAX_TRIES
        else: self.max_reconnect_tries = max_reconnect_tries

        if connected_socket != None:
            self.receives    = True
            self.target_ip   = None
            self.target_port = None
            self.socket      = connected_socket

            if platform.system() == "Linux":
                self.set_timeouts_linux()
            elif platform.system() == "Darwin":
                self.set_timeouts_osx()

            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        elif target_ip != None and target_port != None:
            self.receives    = True
            self.target_ip   = target_ip
            self.target_port = target_port
            self.initiator   = True

            if connect_timeout != None:
                self.connect_timeout = connect_timeout
            else:
                self.connect_timeout = TCPClientInterface.INITIAL_CONNECT_TIMEOUT
            
            if TCPClientInterface.SYNCHRONOUS_START:
                self.initial_connect()
            else:
                thread = threading.Thread(target=self.initial_connect)
                thread.daemon = True
                thread.start()
            
    def initial_connect(self):
        if not self.connect(initial=True):
            thread = threading.Thread(target=self.reconnect)
            thread.daemon = True
            thread.start()
        else:
            thread = threading.Thread(target=self.read_loop)
            thread.daemon = True
            thread.start()
            if not self.kiss_framing:
                self.wants_tunnel = True

    def set_timeouts_linux(self):
        if not self.i2p_tunneled:
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_USER_TIMEOUT, int(TCPClientInterface.TCP_USER_TIMEOUT * 1000))
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, int(TCPClientInterface.TCP_PROBE_AFTER))
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, int(TCPClientInterface.TCP_PROBE_INTERVAL))
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, int(TCPClientInterface.TCP_PROBES))

        else:
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_USER_TIMEOUT, int(TCPClientInterface.I2P_USER_TIMEOUT * 1000))
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, int(TCPClientInterface.I2P_PROBE_AFTER))
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, int(TCPClientInterface.I2P_PROBE_INTERVAL))
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, int(TCPClientInterface.I2P_PROBES))

    def set_timeouts_osx(self):
        if hasattr(socket, "TCP_KEEPALIVE"):
            TCP_KEEPIDLE = socket.TCP_KEEPALIVE
        else:
            TCP_KEEPIDLE = 0x10

        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        
        if not self.i2p_tunneled:
            self.socket.setsockopt(socket.IPPROTO_TCP, TCP_KEEPIDLE, int(TCPClientInterface.TCP_PROBE_AFTER))
        else:
            self.socket.setsockopt(socket.IPPROTO_TCP, TCP_KEEPIDLE, int(TCPClientInterface.I2P_PROBE_AFTER))
        
    def detach(self):
        self.online = False
        if self.socket != None:
            if hasattr(self.socket, "close"):
                if callable(self.socket.close):
                    self.detached = True
                    
                    try:
                        if self.socket != None:
                            self.socket.shutdown(socket.SHUT_RDWR)
                    except Exception as e:
                        RNS.log("Error while shutting down socket for "+str(self)+": "+str(e))

                    try:
                        if self.socket != None:
                            self.socket.close()
                    except Exception as e:
                        RNS.log("Error while closing socket for "+str(self)+": "+str(e))

                    self.socket = None

    def connect(self, initial=False):
        try:
            if initial:
                RNS.log("Establishing TCP connection for "+str(self)+"...", RNS.LOG_DEBUG)

            address_info = socket.getaddrinfo(self.target_ip, self.target_port, proto=socket.IPPROTO_TCP)[0]
            address_family = address_info[0]
            target_address = address_info[4]

            self.socket = socket.socket(address_family, socket.SOCK_STREAM)
            self.socket.settimeout(TCPClientInterface.INITIAL_CONNECT_TIMEOUT)
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            self.socket.connect(target_address)
            self.socket.settimeout(None)
            self.online  = True

            if initial:
                RNS.log("TCP connection for "+str(self)+" established", RNS.LOG_DEBUG)
        
        except Exception as e:
            if initial:
                RNS.log("Initial connection for "+str(self)+" could not be established: "+str(e), RNS.LOG_ERROR)
                RNS.log("Leaving unconnected and retrying connection in "+str(TCPClientInterface.RECONNECT_WAIT)+" seconds.", RNS.LOG_ERROR)
                return False
            
            else:
                raise e

        if platform.system() == "Linux":
            self.set_timeouts_linux()
        elif platform.system() == "Darwin":
            self.set_timeouts_osx()
        
        self.online  = True
        self.writing = False
        self.never_connected = False

        return True


    def reconnect(self):
        if self.initiator:
            if not self.reconnecting:
                self.reconnecting = True
                attempts = 0
                while not self.online:
                    time.sleep(TCPClientInterface.RECONNECT_WAIT)
                    attempts += 1

                    if self.max_reconnect_tries != None and attempts > self.max_reconnect_tries:
                        RNS.log("Max reconnection attempts reached for "+str(self), RNS.LOG_ERROR)
                        self.teardown()
                        break

                    try:
                        self.connect()

                    except Exception as e:
                        RNS.log("Connection attempt for "+str(self)+" failed: "+str(e), RNS.LOG_DEBUG)

                if not self.never_connected:
                    RNS.log("Reconnected socket for "+str(self)+".", RNS.LOG_INFO)

                self.reconnecting = False
                thread = threading.Thread(target=self.read_loop)
                thread.daemon = True
                thread.start()
                if not self.kiss_framing:
                    RNS.Transport.synthesize_tunnel(self)

        else:
            RNS.log("Attempt to reconnect on a non-initiator TCP interface. This should not happen.", RNS.LOG_ERROR)
            raise IOError("Attempt to reconnect on a non-initiator TCP interface")

    def process_incoming(self, data):
        if self.online and not self.detached:
            self.rxb += len(data)
            if hasattr(self, "parent_interface") and self.parent_interface != None:
                self.parent_interface.rxb += len(data)
                        
            self.owner.inbound(data, self)

    def process_outgoing(self, data):
        if self.online and not self.detached:
            # while self.writing:
            #     time.sleep(0.01)

            try:
                self.writing = True

                if self.kiss_framing:
                    data = bytes([KISS.FEND])+bytes([KISS.CMD_DATA])+KISS.escape(data)+bytes([KISS.FEND])
                else:
                    data = bytes([HDLC.FLAG])+HDLC.escape(data)+bytes([HDLC.FLAG])

                self.socket.sendall(data)
                self.writing = False
                self.txb += len(data)
                if hasattr(self, "parent_interface") and self.parent_interface != None:
                    self.parent_interface.txb += len(data)

            except Exception as e:
                RNS.log("Exception occurred while transmitting via "+str(self)+", tearing down interface", RNS.LOG_ERROR)
                RNS.log("The contained exception was: "+str(e), RNS.LOG_ERROR)
                self.teardown()


    def read_loop(self):
        try:
            in_frame = False
            escape = False
            frame_buffer = b""
            data_in = b""
            data_buffer = b""

            while True:
                if self.socket: data_in = self.socket.recv(4096)
                else: data_in = b""
                if len(data_in) > 0:
                    if self.kiss_framing:
                        # Read loop for KISS framing
                        pointer = 0
                        while pointer < len(data_in):
                            byte = data_in[pointer]
                            pointer += 1
                            if (in_frame and byte == KISS.FEND and command == KISS.CMD_DATA):
                                in_frame = False
                                self.process_incoming(data_buffer)
                            elif (byte == KISS.FEND):
                                in_frame = True
                                command = KISS.CMD_UNKNOWN
                                data_buffer = b""
                            elif (in_frame and len(data_buffer) < self.HW_MTU):
                                if (len(data_buffer) == 0 and command == KISS.CMD_UNKNOWN):
                                    # We only support one HDLC port for now, so
                                    # strip off the port nibble
                                    byte = byte & 0x0F
                                    command = byte
                                elif (command == KISS.CMD_DATA):
                                    if (byte == KISS.FESC):
                                        escape = True
                                    else:
                                        if (escape):
                                            if (byte == KISS.TFEND):
                                                byte = KISS.FEND
                                            if (byte == KISS.TFESC):
                                                byte = KISS.FESC
                                            escape = False
                                        data_buffer = data_buffer+bytes([byte])

                    else:
                        # Read loop for standard HDLC framing
                        frame_buffer += data_in
                        flags_remaining = True
                        while flags_remaining:
                            frame_start = frame_buffer.find(HDLC.FLAG)
                            if frame_start != -1:
                                frame_end = frame_buffer.find(HDLC.FLAG, frame_start+1)
                                if frame_end != -1:
                                    frame = frame_buffer[frame_start+1:frame_end]
                                    frame = frame.replace(bytes([HDLC.ESC, HDLC.FLAG ^ HDLC.ESC_MASK]), bytes([HDLC.FLAG]))
                                    frame = frame.replace(bytes([HDLC.ESC, HDLC.ESC  ^ HDLC.ESC_MASK]), bytes([HDLC.ESC]))
                                    if len(frame) > RNS.Reticulum.HEADER_MINSIZE:
                                        self.process_incoming(frame)
                                    frame_buffer = frame_buffer[frame_end:]
                                else:
                                    flags_remaining = False
                            else:
                                flags_remaining = False

                else:
                    self.online = False
                    if self.initiator and not self.detached:
                        RNS.log("The socket for "+str(self)+" was closed, attempting to reconnect...", RNS.LOG_WARNING)
                        self.reconnect()
                    else:
                        RNS.log("The socket for remote client "+str(self)+" was closed.", RNS.LOG_DEBUG)
                        self.teardown()

                    break

                
        except Exception as e:
            self.online = False
            RNS.log("An interface error occurred for "+str(self)+", the contained exception was: "+str(e), RNS.LOG_WARNING)

            if self.initiator:
                RNS.log("Attempting to reconnect...", RNS.LOG_WARNING)
                self.reconnect()
            else:
                self.teardown()

    def teardown(self):
        if self.initiator and not self.detached:
            RNS.log("The interface "+str(self)+" experienced an unrecoverable error and is being torn down. Restart Reticulum to attempt to open this interface again.", RNS.LOG_ERROR)
            if RNS.Reticulum.panic_on_interface_error:
                RNS.panic()

        else:
            RNS.log("The interface "+str(self)+" is being torn down.", RNS.LOG_VERBOSE)

        self.online = False
        self.OUT = False
        self.IN = False

        if hasattr(self, "parent_interface") and self.parent_interface != None:
            while self in self.parent_interface.spawned_interfaces:
                self.parent_interface.spawned_interfaces.remove(self)

        if not self.initiator:
            RNS.Transport.remove_interface(self)


    def __str__(self):
        if ":" in self.target_ip:
            ip_str = f"[{self.target_ip}]"
        else:
            ip_str = f"{self.target_ip}"

        return "TCPInterface["+str(self.name)+"/"+ip_str+":"+str(self.target_port)+"]"


class TCPServerInterface(Interface):
    BITRATE_GUESS     = 10_000_000
    DEFAULT_IFAC_SIZE = 16
    AUTOCONFIGURE_MTU = True

    @staticmethod
    def get_address_for_if(name, bind_port, prefer_ipv6=False):
        from RNS.Interfaces import netinfo
        ifaddr = netinfo.ifaddresses(name)
        if len(ifaddr) < 1:
            raise SystemError(f"No addresses available on specified kernel interface \"{name}\" for TCPServerInterface to bind to")

        if (prefer_ipv6 or not netinfo.AF_INET in ifaddr) and netinfo.AF_INET6 in ifaddr:
            bind_ip = ifaddr[netinfo.AF_INET6][0]["addr"]
            if bind_ip.lower().startswith("fe80::"):
                # We'll need to add the interface as scope for link-local addresses
                return TCPServerInterface.get_address_for_host(f"{bind_ip}%{name}", bind_port, prefer_ipv6)
            else:
                return TCPServerInterface.get_address_for_host(bind_ip, bind_port, prefer_ipv6)
        elif netinfo.AF_INET in ifaddr:
            bind_ip = ifaddr[netinfo.AF_INET][0]["addr"]
            return (bind_ip, bind_port)
        else:
            raise SystemError(f"No addresses available on specified kernel interface \"{name}\" for TCPServerInterface to bind to")

    @staticmethod
    def get_address_for_host(name, bind_port, prefer_ipv6=False):
        address_infos = socket.getaddrinfo(name, bind_port, proto=socket.IPPROTO_TCP)
        address_info  = address_infos[0]
        for entry in address_infos:
            if prefer_ipv6 and entry[0] == socket.AF_INET6:
                address_info = entry; break
            elif not prefer_ipv6 and entry[0] == socket.AF_INET:
                address_info = entry; break

        if address_info[0] == socket.AF_INET6:
            return (name, bind_port, address_info[4][2], address_info[4][3])
        elif address_info[0] == socket.AF_INET:
            return (name, bind_port)
        else:
            raise SystemError(f"No suitable kernel interface available for address \"{name}\" for TCPServerInterface to bind to")


    @property
    def clients(self):
        return len(self.spawned_interfaces)

    def __init__(self, owner, configuration):
        super().__init__()

        c            = Interface.get_config_obj(configuration)
        name         = c["name"]
        device       = c["device"] if "device" in c else None
        port         = int(c["port"]) if "port" in c else None
        bindip       = c["listen_ip"] if "listen_ip" in c else None
        bindport     = int(c["listen_port"]) if "listen_port" in c else None
        i2p_tunneled = c.as_bool("i2p_tunneled") if "i2p_tunneled" in c else False
        prefer_ipv6  = c.as_bool("prefer_ipv6") if "prefer_ipv6" in c else False

        if port != None:
            bindport = port

        self.supports_discovery = True
        self.HW_MTU = TCPInterface.HW_MTU

        self.online = False
        self.spawned_interfaces = []
        
        self.IN  = True
        self.OUT = False
        self.name = name
        self.detached = False

        self.i2p_tunneled = i2p_tunneled
        self.mode         = RNS.Interfaces.Interface.Interface.MODE_FULL

        if bindport == None:
            raise SystemError(f"No TCP port configured for interface \"{name}\"")
        else:
            self.bind_port = bindport

        bind_address = None
        if device != None:
            bind_address = TCPServerInterface.get_address_for_if(device, self.bind_port, prefer_ipv6)
        else:
            if bindip == None:
                raise SystemError(f"No TCP bind IP configured for interface \"{name}\"")
            bind_address = TCPServerInterface.get_address_for_host(bindip, self.bind_port, prefer_ipv6)

        if bind_address != None:
            self.receives = True
            self.bind_ip = bind_address[0]

            def handlerFactory(callback):
                def createHandler(*args, **keys):
                    return TCPInterfaceHandler(callback, *args, **keys)
                return createHandler

            self.owner = owner

            if len(bind_address) == 4:
                try:
                    ThreadingTCP6Server.allow_reuse_address = True
                    self.server = ThreadingTCP6Server(bind_address, handlerFactory(self.incoming_connection))
                except Exception as e:
                    RNS.log(f"Error while binding IPv6 socket for interface, the contained exception was: {e}", RNS.LOG_ERROR)
                    raise SystemError("Could not bind IPv6 socket for interface. Please check the specified \"listen_ip\" configuration option")
            else:
                ThreadingTCPServer.allow_reuse_address = True
                self.server = ThreadingTCPServer(bind_address, handlerFactory(self.incoming_connection))
                self.server.daemon_threads = True

            self.bitrate = TCPServerInterface.BITRATE_GUESS

            thread = threading.Thread(target=self.server.serve_forever)
            thread.daemon = True
            thread.start()

            self.online = True

        else:
            raise SystemError("Insufficient parameters to create TCP listener")

    def incoming_connection(self, handler):
        RNS.log("Accepting incoming TCP connection", RNS.LOG_VERBOSE)
        spawned_configuration = {"name": "Client on "+self.name, "target_host": None, "target_port": None, "i2p_tunneled": self.i2p_tunneled}
        spawned_interface = TCPClientInterface(self.owner, spawned_configuration, connected_socket=handler.request)
        spawned_interface.OUT = self.OUT
        spawned_interface.IN  = self.IN
        
        spawned_interface.ingress_control = self.ingress_control
        spawned_interface.ic_max_held_announces = self.ic_max_held_announces
        spawned_interface.ic_burst_hold = self.ic_burst_hold
        spawned_interface.ic_burst_freq = self.ic_burst_freq
        spawned_interface.ic_burst_freq_new = self.ic_burst_freq_new
        spawned_interface.ic_new_time = self.ic_new_time
        spawned_interface.ic_burst_penalty = self.ic_burst_penalty
        spawned_interface.ic_held_release_interval = self.ic_held_release_interval

        spawned_interface.egress_control = self.egress_control
        spawned_interface.ec_pr_freq = self.ec_pr_freq
        spawned_interface.ic_pr_burst_freq_new = self.ic_pr_burst_freq_new
        spawned_interface.ic_pr_burst_freq = self.ic_pr_burst_freq

        spawned_interface.target_ip = handler.client_address[0]
        spawned_interface.target_port = str(handler.client_address[1])
        spawned_interface.parent_interface = self
        spawned_interface.bitrate = self.bitrate
        spawned_interface.optimise_mtu()
        
        spawned_interface.ifac_size = self.ifac_size
        spawned_interface.ifac_netname = self.ifac_netname
        spawned_interface.ifac_netkey = self.ifac_netkey
        if spawned_interface.ifac_netname != None or spawned_interface.ifac_netkey != None:
            ifac_origin = b""
            if spawned_interface.ifac_netname != None:
                ifac_origin += RNS.Identity.full_hash(spawned_interface.ifac_netname.encode("utf-8"))
            if spawned_interface.ifac_netkey != None:
                ifac_origin += RNS.Identity.full_hash(spawned_interface.ifac_netkey.encode("utf-8"))

            ifac_origin_hash = RNS.Identity.full_hash(ifac_origin)
            spawned_interface.ifac_key = RNS.Cryptography.hkdf(
                length=64,
                derive_from=ifac_origin_hash,
                salt=RNS.Reticulum.IFAC_SALT,
                context=None
            )
            spawned_interface.ifac_identity = RNS.Identity.from_bytes(spawned_interface.ifac_key)
            spawned_interface.ifac_signature = spawned_interface.ifac_identity.sign(RNS.Identity.full_hash(spawned_interface.ifac_key))

        spawned_interface.announce_rate_target = self.announce_rate_target
        spawned_interface.announce_rate_grace = self.announce_rate_grace
        spawned_interface.announce_rate_penalty = self.announce_rate_penalty
        spawned_interface.mode = self.mode
        spawned_interface.HW_MTU = self.HW_MTU
        spawned_interface.online = True
        RNS.log("Spawned new TCPClient Interface: "+str(spawned_interface), RNS.LOG_VERBOSE)
        RNS.Transport.add_interface(spawned_interface)
        while spawned_interface in self.spawned_interfaces:
            self.spawned_interfaces.remove(spawned_interface)
        self.spawned_interfaces.append(spawned_interface)
        spawned_interface.read_loop()

    def received_announce(self, from_spawned=False):
        if from_spawned: self.ia_freq_deque.append(time.time())

    def sent_announce(self, from_spawned=False):
        if from_spawned: self.oa_freq_deque.append(time.time())

    def received_path_request(self, from_spawned=False):
        if from_spawned: self.ip_freq_deque.append(time.time())

    def sent_path_request(self, from_spawned=False):
        if from_spawned: self.op_freq_deque.append(time.time())

    def process_outgoing(self, data):
        pass

    def detach(self):
        self.detached = True
        self.online = False
        if self.server != None:
            if hasattr(self.server, "shutdown"):
                if callable(self.server.shutdown):
                    try:
                        RNS.log("Detaching "+str(self), RNS.LOG_DEBUG)
                        self.server.shutdown()
                        self.server.server_close()
                        self.server = None

                    except Exception as e:
                        RNS.log("Error while shutting down server for "+str(self)+": "+str(e))


    def __str__(self):
        if ":" in self.bind_ip:
            ip_str = f"[{self.bind_ip}]"
        else:
            ip_str = f"{self.bind_ip}"

        return "TCPServerInterface["+self.name+"/"+ip_str+":"+str(self.bind_port)+"]"


class TCPInterfaceHandler(socketserver.BaseRequestHandler):
    def __init__(self, callback, *args, **keys):
        self.callback = callback
        socketserver.BaseRequestHandler.__init__(self, *args, **keys)

    def handle(self):
        self.callback(handler=self)



compiling = False
noticed = False
notice_delay = 0.3
import time
import sys
import threading
from importlib.util import find_spec
if find_spec("pyximport") and find_spec("cython"):
    import pyximport; pyxloader = pyximport.install(pyimport=True, language_level=3)[1]

def notice_job():
    global noticed
    started = time.time()
    while compiling:
        if time.time() > started+notice_delay and compiling:
            noticed = True
            print("Compiling RNS object code... ", end="")
            sys.stdout.flush()
            break
        time.sleep(0.1)


compiling = True
threading.Thread(target=notice_job, daemon=True).start()
import RNS; compiling = False
if noticed: print("Done."); sys.stdout.flush()


# MIT License
#
# Copyright (c) 2014 Stefan C. Mueller
# Copyright (c) 2025 Mark Qvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import socket
import ipaddress
import platform
import ctypes.util
import collections
from typing import List, Iterable, Optional, Tuple, Union

AF_INET6 = socket.AF_INET6.value
AF_INET = socket.AF_INET.value

def interfaces() -> List[str]:
    adapters = get_adapters(include_unconfigured=True)
    return [a.name for a in adapters]

def interface_names_to_indexes() -> dict:
    adapters = get_adapters(include_unconfigured=True)
    results = {}
    for adapter in adapters:
        results[adapter.name] = adapter.index
    return results

def interface_name_to_nice_name(ifname) -> str:
    try:
        adapters = get_adapters(include_unconfigured=True)
        for adapter in adapters:
            if adapter.name == ifname:
                if hasattr(adapter, "nice_name"):
                    return adapter.nice_name

    except: return None
    return None

def ifaddresses(ifname) -> dict:
    adapters = get_adapters(include_unconfigured=True)
    ifa = {}
    for a in adapters:
        if a.name == ifname:
            ipv4s = []
            ipv6s = []
            for ip in a.ips:
                t = {}
                if ip.is_IPv4:
                    net = ipaddress.ip_network(str(ip.ip)+"/"+str(ip.network_prefix), strict=False)
                    t["addr"] = ip.ip
                    t["prefix"] = ip.network_prefix
                    t["broadcast"] = str(net.broadcast_address)
                    ipv4s.append(t)
                if ip.is_IPv6:
                    t["addr"] = ip.ip[0]
                    ipv6s.append(t)

            if len(ipv4s) > 0: ifa[AF_INET] = ipv4s
            if len(ipv6s) > 0: ifa[AF_INET6] = ipv6s

    return ifa

def get_adapters(include_unconfigured=False):
    if os.name == "posix": return _get_adapters_posix(include_unconfigured=include_unconfigured)
    elif os.name == "nt":  return _get_adapters_win(include_unconfigured=include_unconfigured)
    else: raise RuntimeError(f"Unsupported Operating System: {os.name}")

class Adapter(object):
    def __init__(self, name: str, nice_name: str, ips: List["IP"], index: Optional[int] = None) -> None:
        self.name = name
        self.nice_name = nice_name
        self.ips = ips
        self.index = index

    def __repr__(self) -> str:
        return "Adapter(name={name}, nice_name={nice_name}, ips={ips}, index={index})".format(
            name=repr(self.name), nice_name=repr(self.nice_name), ips=repr(self.ips), index=repr(self.index))

_IPv4Address = str
_IPv6Address = Tuple[str, int, int]
class IP(object):
    def __init__(self, ip: Union[_IPv4Address, _IPv6Address], network_prefix: int, nice_name: str) -> None:
        self.ip = ip
        self.network_prefix = network_prefix
        self.nice_name = nice_name

    @property
    def is_IPv4(self) -> bool: return not isinstance(self.ip, tuple)

    @property
    def is_IPv6(self) -> bool: return isinstance(self.ip, tuple)

    def __repr__(self) -> str:
        return "IP(ip={ip}, network_prefix={network_prefix}, nice_name={nice_name})".format(ip=repr(self.ip), network_prefix=repr(self.network_prefix), nice_name=repr(self.nice_name))

if platform.system() == "Darwin" or "BSD" in platform.system():
    class sockaddr(ctypes.Structure):
        _fields_ = [
            ("sa_len", ctypes.c_uint8),
            ("sa_familiy", ctypes.c_uint8),
            ("sa_data", ctypes.c_uint8 * 14)]

    class sockaddr_in(ctypes.Structure):
        _fields_ = [
            ("sa_len", ctypes.c_uint8),
            ("sa_familiy", ctypes.c_uint8),
            ("sin_port", ctypes.c_uint16),
            ("sin_addr", ctypes.c_uint8 * 4),
            ("sin_zero", ctypes.c_uint8 * 8)]

    class sockaddr_in6(ctypes.Structure):
        _fields_ = [
            ("sa_len", ctypes.c_uint8),
            ("sa_familiy", ctypes.c_uint8),
            ("sin6_port", ctypes.c_uint16),
            ("sin6_flowinfo", ctypes.c_uint32),
            ("sin6_addr", ctypes.c_uint8 * 16),
            ("sin6_scope_id", ctypes.c_uint32)]

else:
    class sockaddr(ctypes.Structure):  # type: ignore
        _fields_ = [("sa_familiy", ctypes.c_uint16), ("sa_data", ctypes.c_uint8 * 14)]

    class sockaddr_in(ctypes.Structure):  # type: ignore
        _fields_ = [
            ("sin_familiy", ctypes.c_uint16),
            ("sin_port", ctypes.c_uint16),
            ("sin_addr", ctypes.c_uint8 * 4),
            ("sin_zero", ctypes.c_uint8 * 8)]

    class sockaddr_in6(ctypes.Structure):  # type: ignore
        _fields_ = [
            ("sin6_familiy", ctypes.c_uint16),
            ("sin6_port", ctypes.c_uint16),
            ("sin6_flowinfo", ctypes.c_uint32),
            ("sin6_addr", ctypes.c_uint8 * 16),
            ("sin6_scope_id", ctypes.c_uint32)]

def sockaddr_to_ip(sockaddr_ptr: "ctypes.pointer[sockaddr]") -> Optional[Union[_IPv4Address, _IPv6Address]]:
    if sockaddr_ptr:
        if sockaddr_ptr[0].sa_familiy == socket.AF_INET:
            ipv4 = ctypes.cast(sockaddr_ptr, ctypes.POINTER(sockaddr_in))
            ippacked = bytes(bytearray(ipv4[0].sin_addr))
            ip = str(ipaddress.ip_address(ippacked))
            return ip
        elif sockaddr_ptr[0].sa_familiy == socket.AF_INET6:
            ipv6 = ctypes.cast(sockaddr_ptr, ctypes.POINTER(sockaddr_in6))
            flowinfo = ipv6[0].sin6_flowinfo
            ippacked = bytes(bytearray(ipv6[0].sin6_addr))
            ip = str(ipaddress.ip_address(ippacked))
            scope_id = ipv6[0].sin6_scope_id
            return (ip, flowinfo, scope_id)
    return None


def ipv6_prefixlength(address: ipaddress.IPv6Address) -> int:
    prefix_length = 0
    for i in range(address.max_prefixlen):
        if int(address) >> i & 1: prefix_length = prefix_length + 1
    return prefix_length

if os.name == "posix":
    class ifaddrs(ctypes.Structure): pass
    ifaddrs._fields_ = [
        ("ifa_next", ctypes.POINTER(ifaddrs)),
        ("ifa_name", ctypes.c_char_p),
        ("ifa_flags", ctypes.c_uint),
        ("ifa_addr", ctypes.POINTER(sockaddr)),
        ("ifa_netmask", ctypes.POINTER(sockaddr)),]

    libc = ctypes.CDLL(ctypes.util.find_library("socket" if os.uname()[0] == "SunOS" else "c"), use_errno=True) # type: ignore

    def _get_adapters_posix(include_unconfigured: bool = False) -> Iterable[Adapter]:
        addr0 = addr = ctypes.POINTER(ifaddrs)()
        retval = libc.getifaddrs(ctypes.byref(addr))
        if retval != 0:
            eno = ctypes.get_errno()
            raise OSError(eno, os.strerror(eno))

        ips = collections.OrderedDict()

        def add_ip(adapter_name: str, ip: Optional[IP]) -> None:
            if adapter_name not in ips:
                index = None  # type: Optional[int]
                try:
                    index = socket.if_nametoindex(adapter_name) # type: ignore
                except (OSError, AttributeError): pass
                ips[adapter_name] = Adapter(adapter_name, adapter_name, [], index=index)
            if ip is not None:
                ips[adapter_name].ips.append(ip)

        while addr:
            name = addr[0].ifa_name.decode(encoding="UTF-8")
            ip_addr = sockaddr_to_ip(addr[0].ifa_addr)
            if ip_addr:
                if addr[0].ifa_netmask and not addr[0].ifa_netmask[0].sa_familiy:
                    addr[0].ifa_netmask[0].sa_familiy = addr[0].ifa_addr[0].sa_familiy
                netmask = sockaddr_to_ip(addr[0].ifa_netmask)
                if isinstance(netmask, tuple):
                    netmaskStr = str(netmask[0])
                    prefixlen = ipv6_prefixlength(ipaddress.IPv6Address(netmaskStr))
                else:
                    assert netmask is not None, f"sockaddr_to_ip({addr[0].ifa_netmask}) returned None"
                    netmaskStr = str("0.0.0.0/" + netmask)
                    prefixlen = ipaddress.IPv4Network(netmaskStr).prefixlen
                ip = IP(ip_addr, prefixlen, name)
                add_ip(name, ip)
            else:
                if include_unconfigured:
                    add_ip(name, None)
            addr = addr[0].ifa_next

        libc.freeifaddrs(addr0)
        return ips.values()

elif os.name == "nt":
    from ctypes import wintypes
    NO_ERROR = 0
    ERROR_BUFFER_OVERFLOW = 111
    MAX_ADAPTER_NAME_LENGTH = 256
    MAX_ADAPTER_DESCRIPTION_LENGTH = 128
    MAX_ADAPTER_ADDRESS_LENGTH = 8
    AF_UNSPEC = 0

    class SOCKET_ADDRESS(ctypes.Structure): _fields_ = [("lpSockaddr", ctypes.POINTER(sockaddr)), ("iSockaddrLength", wintypes.INT)]
    class IP_ADAPTER_UNICAST_ADDRESS(ctypes.Structure): pass
    IP_ADAPTER_UNICAST_ADDRESS._fields_ = [
        ("Length", wintypes.ULONG),
        ("Flags", wintypes.DWORD),
        ("Next", ctypes.POINTER(IP_ADAPTER_UNICAST_ADDRESS)),
        ("Address", SOCKET_ADDRESS),
        ("PrefixOrigin", ctypes.c_uint),
        ("SuffixOrigin", ctypes.c_uint),
        ("DadState", ctypes.c_uint),
        ("ValidLifetime", wintypes.ULONG),
        ("PreferredLifetime", wintypes.ULONG),
        ("LeaseLifetime", wintypes.ULONG),
        ("OnLinkPrefixLength", ctypes.c_uint8)]

    class IP_ADAPTER_ADDRESSES(ctypes.Structure): pass
    IP_ADAPTER_ADDRESSES._fields_ = [
        ("Length", wintypes.ULONG),
        ("IfIndex", wintypes.DWORD),
        ("Next", ctypes.POINTER(IP_ADAPTER_ADDRESSES)),
        ("AdapterName", ctypes.c_char_p),
        ("FirstUnicastAddress", ctypes.POINTER(IP_ADAPTER_UNICAST_ADDRESS)),
        ("FirstAnycastAddress", ctypes.c_void_p),
        ("FirstMulticastAddress", ctypes.c_void_p),
        ("FirstDnsServerAddress", ctypes.c_void_p),
        ("DnsSuffix", ctypes.c_wchar_p),
        ("Description", ctypes.c_wchar_p),
        ("FriendlyName", ctypes.c_wchar_p)]

    iphlpapi = ctypes.windll.LoadLibrary("Iphlpapi") # type: ignore

    def _enumerate_interfaces_of_adapter_win(nice_name: str, address: IP_ADAPTER_UNICAST_ADDRESS) -> Iterable[IP]:
        # Iterate through linked list and fill list
        addresses = [] # type: List[IP_ADAPTER_UNICAST_ADDRESS]
        while True:
            addresses.append(address)
            if not address.Next: break
            address = address.Next[0]

        for address in addresses:
            ip = sockaddr_to_ip(address.Address.lpSockaddr)
            assert ip is not None, f"sockaddr_to_ip({address.Address.lpSockaddr}) returned None"
            network_prefix = address.OnLinkPrefixLength
            yield IP(ip, network_prefix, nice_name)

    def _get_adapters_win(include_unconfigured: bool = False) -> Iterable[Adapter]:
        addressbuffersize = wintypes.ULONG(15 * 1024)
        retval = ERROR_BUFFER_OVERFLOW
        while retval == ERROR_BUFFER_OVERFLOW:
            addressbuffer = ctypes.create_string_buffer(addressbuffersize.value)
            retval = iphlpapi.GetAdaptersAddresses(
                wintypes.ULONG(AF_UNSPEC),
                wintypes.ULONG(0),
                None,
                ctypes.byref(addressbuffer),
                ctypes.byref(addressbuffersize))

        if retval != NO_ERROR:
            raise ctypes.WinError() # type: ignore

        # Iterate through adapters and fill array
        address_infos = []  # type: List[IP_ADAPTER_ADDRESSES]
        address_info = IP_ADAPTER_ADDRESSES.from_buffer(addressbuffer)
        while True:
            address_infos.append(address_info)
            if not address_info.Next: break
            address_info = address_info.Next[0]

        # Iterate through unicast addresses
        result = [] # type: List[Adapter]
        for adapter_info in address_infos:
            name = adapter_info.AdapterName.decode()
            nice_name = adapter_info.Description
            index = adapter_info.IfIndex

            if adapter_info.FirstUnicastAddress:
                ips = _enumerate_interfaces_of_adapter_win(adapter_info.FriendlyName, adapter_info.FirstUnicastAddress[0])
                ips = list(ips)
                result.append(Adapter(name, nice_name, ips, index=index))
            
            elif include_unconfigured: result.append(Adapter(name, nice_name, [], index=index))

        return result


# Reticulum License
#
# Copyright (c) 2016-2026 Mark Qvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# - The Software shall not be used in any kind of system which includes amongst
#   its functions the ability to purposefully do harm to human beings.
#
# - The Software shall not be used, directly or indirectly, in the creation of
#   an artificial intelligence, machine learning or language model training
#   dataset, including but not limited to any use that contributes to the
#   training or development of such a model or algorithm.
#
# - The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import io
import RNS

class SyntaxHighlighter:
    
    def __init__(self, theme=None):
        self.pygments_available = False
        self.pygments = None
        self._lexer_cache = {}
        self._check_pygments()
        self.theme = theme or self._get_default_theme()
    
    def _get_default_theme(self):
        return {
            # Control flow - warm coral-red
            "keyword": "ff7b72",
            "keyword_constant": "ff7b72",
            "keyword_control": "ff7b72",
            "keyword_declaration": "ff7b72",
            
            # Function definitions - bright sky blue
            "function_def": "79c0ff",
            "function_magic": "ff7b72",
            
            # Function calls - soft lavender
            "function_call": "d2a8ff",
            "function_builtin": "ffa657",    # amber
            
            # Class definitions - fresh mint green
            "class_def": "7ee787",
            "class_ref": "56d364",           # muted when referenced
            
            # Instance context - soft pink
            "self": "ff9bce",
            "cls": "ff9bce",
            
            # Data literals - cool, calm ice blue
            "string": "a5d6ff",
            "string_quoted": "a5d6ff",
            "string_doc": "8b949e",          # docstrings - like comments
            "string_interpol": "ffd700",     # f-string braces - gold
            "string_escape": "ffea00",       # escape sequences - bright yellow
            
            # Numbers - same as function def
            "number": "79c0ff",
            "number_float": "79c0ff",
            "number_integer": "79c0ff",
            "number_hex": "79c0ff",
            
            # Comments - muted gray
            "comment": "8b949e",
            "comment_doc": "8b949e",
            "comment_preproc": "ff7b72",     # preprocessor directives
            
            # Operators - distinct pink/red for visibility
            "operator": "ff7b72",            # General operators - coral
            "operator_arithmetic": "ff7b72", # +, -, *, /, etc.
            "operator_comparison": "ff7b72", # ==, !=, <, >, etc.
            "operator_assignment": "ff7b72", # =, +=, -=, etc.
            "operator_word": "ff7b72",       # and, or, not, in, is
            "operator_dot": "c9d1d9",        # . - subtle for attribute access
            
            # Punctuation - neutral
            "punctuation": "b4b4b4",
            "punctuation_brace": "b4b4b4",   # [, ], {, }
            "punctuation_paren": "b4b4b4",   # (, )
            "punctuation_colon": "b4b4b4",   # :, ;
            "punctuation_comma": "8b949e",   # , - slightly dimmed
            
            # Decorators - burnt orange
            "decorator": "f0883e",
            
            # Constants - same as keywords
            "constant": "ff7b72",
            "constant_builtin": "ff7b72",    # True, False, None
            
            # Type hints and annotations - amber
            "type_hint": "ffa657",
            "type_builtin": "ffa657",
            
            # Exception handling - alert red
            "exception": "f85149",
            "exception_builtin": "f85149",
            
            # Names and attributes - near-white for readability
            "name": "e6edf3",
            "attribute": "e6edf3",
            "attribute_call": "d2a8ff",       # Function/method calls after dot - lavender
            "variable": "e6edf3",
            "parameter": "e6edf3",
            
            # Namespaces and modules
            "namespace": "7ee787",
            "module": "a5d6ff",
            
            # Generic tokens
            "generic_heading": "c9d1d9",
            "generic_subheading": "c9d1d9",
            "generic_prompt": "8b949e",
            "generic_error": "f85149",
            "generic_deleted": "f85149",
            "generic_inserted": "7ee787",
            "generic_output": "e6edf3",
            
            # Text and whitespace - no color (None means no color tag)
            "text": None,
            "whitespace": None,
        }
    
    def _check_pygments(self):
        try:
            import pygments
            from pygments.lexers import get_lexer_for_filename, guess_lexer, get_lexer_by_name
            from pygments.formatter import Formatter
            from pygments.token import Token
            
            self.pygments = pygments
            self.pygments_available = True
            RNS.log("Pygments syntax highlighting available", RNS.LOG_DEBUG)
            
        except ImportError:
            self.pygments_available = False
            RNS.log("Pygments not available, using plain text rendering", RNS.LOG_DEBUG)
    
    def highlight(self, content, filename=None, language=None):
        if not content: return self._plain_text(content)
        
        if self.pygments_available:
            try:
                highlighted = self._highlight_pygments(content, filename, language)
                # Fix pygments insisting on trailing newlines
                if highlighted.endswith("\n") and not content.endswith("\n"): highlighted = highlighted[:-1]
                return highlighted
            
            except Exception as e:
                RNS.log(f"Pygments highlighting failed, falling back: {e}", RNS.LOG_WARNING)
                return self._plain_text(content).replace("\\", "\\\\")
        
        # TODO: Implement Python tokenize fallback for .py files.
        # For now, route to plain text
        if filename and filename.endswith(".py"):
            return self._plain_text(content).replace("\\", "\\\\")
        
        # Universal fallback
        return self._plain_text(content).replace("\\", "\\\\")
    
    def _highlight_pygments(self, content, filename=None, language=None):
        from pygments.lexers import get_lexer_for_filename, guess_lexer, get_lexer_by_name
        from pygments.util import ClassNotFound
        
        lexer = None        
        if language:
            if language == "env":         language = "bash"
            if language == "environment": language = "bash"
            try: lexer = get_lexer_by_name(language)
            except ClassNotFound: pass
        
        if lexer is None and filename:
            try: lexer = get_lexer_for_filename(filename)
            except ClassNotFound: pass
        
        if lexer is None:
            try:
                if len(content) > 20: lexer = guess_lexer(content)
            except ClassNotFound: pass
        
        if lexer is None: return self._plain_text(content)

        formatter = MicronFormatter(theme=self.theme)
        result = self.pygments.highlight(content, lexer, formatter)
        return result
    
    def _plain_text(self, content):
        escaped = self._escape_micron(content)
        return f"`=\n{escaped}\n`="
    
    @staticmethod
    def _escape_micron(text): return text.replace("`", "\\`")


class MicronFormatter:
    def __init__(self, theme, **options):
        self.theme = theme
        self.options = options
    
    def format(self, tokensource, outfile):
        output_parts = []
        prev_was_dot = False
        
        last_ended_with_break = True
        for ttype, value in tokensource:
            is_dot = (str(ttype) == "Token.Operator" and value == ".")
            ends_with_break = value.endswith("\n")
            
            # If previous token was a dot and this is a Name, treat as attribute/function call
            # TODO: Improve this if we can check next token as parantheses or something.
            if prev_was_dot and str(ttype).startswith("Token.Name") and value:
                color = self._get_color_from_key("attribute_call")
                if color:
                    escaped = self._escape_value(value)
                    output_parts.append(f"`FT{color}{escaped}`f")
                else:
                    output_parts.append(self._escape_value(value))
            
            else:
                color_key = self._get_color_key_for_token(ttype)
                color = self._get_color_from_key(color_key)
                
                if color and value:
                    escaped = self._escape_value(value)
                    if escaped.startswith("\n"): ilb = "\n"; escaped = escaped[1:]
                    else:                        ilb = ""
                    if escaped.endswith("\n"):   tlb = "\n"; escaped = escaped[:-1]
                    else:                        tlb = ""

                    if len(escaped): output = f"{ilb}`FT{color}{escaped}`f{tlb}"
                    else:            output = f"{ilb}{tlb}"

                    output_parts.append(output)
                
                else:
                    escaped = self._escape_value(value)
                    if "\n" in escaped:
                        parts = []
                        splitl = escaped.splitlines()
                        if len(splitl) > 1:
                            for line in splitl:
                                if   line.startswith("-"): l = f"\\{line}"
                                elif line.startswith(">"): l = f"\\{line}"
                                elif line.startswith("<"): l = f"\\{line}"
                                else:                      l = line
                                parts.append(l)
                            trmpart = "\n" if escaped.endswith("\n") else ""
                            escaped = "\n".join(parts)+trmpart

                    elif last_ended_with_break:
                        if   escaped.startswith("-"): escaped = f"\\{escaped}"
                        elif escaped.startswith(">"): escaped = f"\\{escaped}"
                        elif escaped.startswith("<"): escaped = f"\\{escaped}"
                    
                    output_parts.append(escaped)
            
            prev_was_dot = is_dot
            last_ended_with_break = ends_with_break
        
        output = "".join(output_parts)
        outfile.write(output)
    
    def _get_color_key_for_token(self, ttype):
        token_parts = []
        current = ttype
        while current:
            token_parts.insert(0, current[0] if isinstance(current, tuple) else str(current).split(".")[-1])
            current = current.parent if hasattr(current, "parent") else None
        
        token_str = ".".join(["Token"] + token_parts[1:] if len(token_parts) > 1 else token_parts)

        current_type = ttype
        while current_type:
            token_key = str(current_type)
            if token_key in granular_token_map: return granular_token_map[token_key]
            
            # Move to parent
            current_type = current_type.parent if hasattr(current_type, "parent") else None
        
        return None
    
    def _get_color_from_key(self, color_key):
        if color_key and color_key in self.theme: return self.theme[color_key]
        return None
    
    @staticmethod
    def _escape_value(value):
        return value.replace("\\", "\\\\").replace("`", "\\`")
    
    # Required by Pygments formatter API, returns None for Micron
    def get_style_defs(self, arg=None): return None


# Convenience function for direct use
def highlight_code(content: str, filename: str = None, language: str = None, theme=None) -> str:
    highlighter = SyntaxHighlighter(theme=theme)
    return highlighter.highlight(content, filename, language)

granular_token_map = {
    # Keywords with semantic distinction
    "Token.Keyword": "keyword",
    "Token.Keyword.Constant": "keyword_constant",
    "Token.Keyword.Declaration": "keyword_declaration",
    "Token.Keyword.Namespace": "keyword_control",
    "Token.Keyword.Pseudo": "keyword_control",
    "Token.Keyword.Reserved": "keyword_control",
    "Token.Keyword.Type": "type_builtin",
    
    # Names - functions with definition vs call distinction
    "Token.Name.Function": "function_call",
    "Token.Name.Function.Magic": "function_magic",
    "Token.Name.Class": "class_ref",
    "Token.Name.Builtin": "function_builtin",
    "Token.Name.Builtin.Pseudo": "constant_builtin",
    "Token.Name.Exception": "exception_builtin",
    "Token.Name.Decorator": "decorator",
    "Token.Name.Namespace": "namespace",
    "Token.Name.Attribute": "attribute",
    "Token.Name.Variable": "variable",
    "Token.Name.Variable.Magic": "function_magic",
    "Token.Name.Other": "name",
    "Token.Name": "name",
    "Token.Name.Tag": "keyword",  # HTML/XML tags
    "Token.Name.Constant": "constant",
    "Token.Name.Label": "name",
    "Token.Name.Entity": "name",
    
    # Literals - strings with detailed handling
    "Token.Literal.String": "string",
    "Token.Literal.String.Affix": "string",  # f, r, b prefixes
    "Token.Literal.String.Backtick": "string",
    "Token.Literal.String.Char": "string",
    "Token.Literal.String.Delimiter": "string",
    "Token.Literal.String.Doc": "string_doc",
    "Token.Literal.String.Double": "string_quoted",
    "Token.Literal.String.Escape": "string_escape",
    "Token.Literal.String.Heredoc": "string",
    "Token.Literal.String.Interpol": "string_interpol",
    "Token.Literal.String.Other": "string",
    "Token.Literal.String.Regex": "string",
    "Token.Literal.String.Single": "string_quoted",
    "Token.Literal.String.Symbol": "string",
    
    # Numbers
    "Token.Literal.Number": "number",
    "Token.Literal.Number.Bin": "number",
    "Token.Literal.Number.Float": "number_float",
    "Token.Literal.Number.Hex": "number_hex",
    "Token.Literal.Number.Integer": "number_integer",
    "Token.Literal.Number.Integer.Long": "number_integer",
    "Token.Literal.Number.Oct": "number",
    "Token.Literal": "string",
    "Token.Literal.Date": "string",
    
    # Operators - all operators get distinct coloring
    "Token.Operator": "operator",
    "Token.Operator.Word": "operator_word",
    "Token.Operator.Comparison": "operator_comparison",
    "Token.Operator.Assignment": "operator_assignment",
    "Token.Operator.Arithmetic": "operator_arithmetic",
    
    # Punctuation - braces, parens, colons, commas
    "Token.Punctuation": "punctuation",
    "Token.Punctuation.Marker": "punctuation",
    "Token.Punctuation.Brace": "punctuation_brace",
    "Token.Punctuation.Bracket": "punctuation_brace",
    "Token.Punctuation.Parenthesis": "punctuation_paren",
    "Token.Punctuation.Colon": "punctuation_colon",
    "Token.Punctuation.Comma": "punctuation_comma",
    
    # Comments
    "Token.Comment": "comment",
    "Token.Comment.Hashbang": "comment",
    "Token.Comment.Multiline": "comment_doc",
    "Token.Comment.Preproc": "comment_preproc",
    "Token.Comment.Single": "comment",
    "Token.Comment.Special": "comment",
    
    # Generic tokens
    "Token.Generic.Deleted": "generic_deleted",
    "Token.Generic.Emph": "text",
    "Token.Generic.Error": "generic_error",
    "Token.Generic.Heading": "generic_heading",
    "Token.Generic.Inserted": "generic_inserted",
    "Token.Generic.Output": "generic_output",
    "Token.Generic.Prompt": "generic_prompt",
    "Token.Generic.Strong": "text",
    "Token.Generic.Subheading": "generic_subheading",
    "Token.Generic.Traceback": "generic_error",
    "Token.Generic": "text",
    
    # Text and whitespace
    "Token.Text": "text",
    "Token.Text.Whitespace": "whitespace",
}


#!/usr/bin/env python3

# Reticulum License
#
# Copyright (c) 2016-2026 Mark Qvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# - The Software shall not be used in any kind of system which includes amongst
#   its functions the ability to purposefully do harm to human beings.
#
# - The Software shall not be used, directly or indirectly, in the creation of
#   an artificial intelligence, machine learning or language model training
#   dataset, including but not limited to any use that contributes to the
#   training or development of such a model or algorithm.
#
# - The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import RNS
import os
import sys
import time
import shutil
import threading
import subprocess

from RNS._version import __version__
from RNS.Utilities.rngit import APP_NAME

from RNS.vendor.configobj import ConfigObj
from tempfile import TemporaryDirectory

def program_setup(configdir, rnsconfigdir, destination_hexhash, group_name, repo_name):
    git_client = ReticulumGitClient(configdir=configdir, rnsconfigdir=rnsconfigdir, destination_hexhash=destination_hexhash,
                                    group_name=group_name, repo_name=repo_name)

    if not git_client.ready: sys.exit(1)
    else:                    git_client.run()

def main():
    if len(sys.argv) < 3:
        print("Usage: git-remote-rns <remote-name> <url>", file=sys.stderr)
        sys.exit(1)
    
    url = sys.argv[2]
    if not url.startswith("rns://"):
        print("Invalid URL scheme. Must be rns://", file=sys.stderr)
        sys.exit(1)

    try:
        parts = url[6:].split("/", 2)
        destination_hexhash = parts[0]
        group_name = parts[1]
        repo_name = parts[2]
    
    except IndexError: print("Invalid URL format. Use rns://<hash>/<group>/<repo>", file=sys.stderr); sys.exit(1)

    configdir    = os.environ.get("RNGIT_CONFIG", None)
    rnsconfigdir = os.environ.get("RNS_CONFIG", None)

    program_setup(configdir, rnsconfigdir, destination_hexhash, group_name, repo_name)
    exit(0)


class ReticulumGitClient():
    PATH_LIST       = "/git/list"
    PATH_FETCH      = "/git/fetch"
    PATH_PUSH       = "/git/push"
    PATH_DELETE     = "/git/delete"

    RES_DISALLOWED  = 0x01
    RES_INVALID_REQ = 0x02
    RES_NOT_FOUND   = 0x03
    RES_REMOTE_FAIL = 0xFF

    IDX_REPOSITORY  = 0x00
    IDX_RESULT_CODE = 0x01

    REF_BATCH_SIZE  = 25
    PATH_TIMEOUT    = 15
    LINK_TIMEOUT    = 15

    def __init__(self, configdir, rnsconfigdir, destination_hexhash, group_name, repo_name):
        # Client state and configuration
        self.identity            = None
        self.userdir             = os.path.expanduser("~")
        self.config              = None
        self.ready               = False
        
        self.destination_aliases = {}
        self.remote_identity     = None
        self.destination         = None
        self.link                = None
        self.link_ready          = False
        self.link_failed         = False
        self.link_timeout        = self.LINK_TIMEOUT
        self.path_timeout        = self.PATH_TIMEOUT

        self.destination_hexhash = destination_hexhash
        self.group_name          = group_name
        self.repo_name           = repo_name
        self.repo_path           = f"{group_name}/{repo_name}"

        self.tmp_dir             = TemporaryDirectory()
        self.request_event       = threading.Event()
        self.request_response    = None
        self.response_metadata   = None

        self.ref_batch_size      = self.REF_BATCH_SIZE
        self.remote_refs         = {}

        self.response_progress      = 0
        self.previous_progress      = 0
        self.response_size          = None
        self.response_transfer_size = None
        self.progress_updated_at    = None
        self.progress_enabled       = False

        if configdir != None: self.configdir = configdir
        else:
            if os.path.isdir(self.userdir+"/.config/rngit") and os.path.isfile(self.userdir+"/.config/rngit/config"): self.configdir = self.userdir+"/.rngit/reticulum"
            else: self.configdir = self.userdir+"/.rngit"
        
        self.logfile = self.configdir+"/client_log"
        self.configpath = self.configdir+"/client_config"
        self.identitypath = self.configdir+"/client_identity"

        if os.path.isfile(self.configpath):
            try: self.config = ConfigObj(self.configpath)
            except Exception as e:
                RNS.log("Could not parse the configuration at "+self.configpath, RNS.LOG_ERROR)
                return
        
        else: self.__create_default_config()

        RNS.logfile = self.logfile
        try: self.reticulum = RNS.Reticulum(configdir=rnsconfigdir, logdest=RNS.LOG_FILE)
        except Exception as e:
            print(f"Failed to initialize Reticulum: {e}", file=sys.stderr)
            return

        self.__apply_config()
        self.ready = True

    def __create_default_config(self):
        self.config = ConfigObj(__default_rngit_config__)
        self.config.filename = self.configpath
        if not os.path.isdir(self.configdir): os.makedirs(self.configdir)
        self.config.write()

    def __apply_config(self):
        if "logging" in self.config:
            section = self.config["logging"]
            if "loglevel" in section: RNS.loglevel = max(RNS.LOG_NONE, min(RNS.LOG_EXTREME, section.as_int("loglevel")))
        
        if "client" in self.config:
            section = self.config["client"]
            if "ref_batch_size" in section: self.ref_batch_size = max(0, min(1024, section.as_int("ref_batch_size")))

        if "aliases" in self.config:
            section = self.config["aliases"]
            for alias in section:
                alias_hexhash = section[alias]
                len_ok = len(alias_hexhash) == RNS.Identity.TRUNCATED_HASHLENGTH//8*2
                try: alias_hash = bytes.fromhex(alias_hexhash)
                except: alias_hash = None
                alias_exists = alias in self.destination_aliases
                if not len_ok or not alias_hash: continue
                if alias_exists: continue
                self.destination_aliases[alias] = RNS.hexrep(alias_hash, delimit=False)

        if not os.path.isfile(self.identitypath):
            identity = RNS.Identity()
            identity.to_file(self.identitypath)
            RNS.log(f"Client identity generated and persisted to {self.identitypath}", RNS.LOG_VERBOSE)
        
        else:
            identity = RNS.Identity.from_file(self.identitypath)
            RNS.log(f"Client identity loaded from {self.identitypath}", RNS.LOG_VERBOSE)

        if not identity:
            RNS.log("Could not initialize client identity.", RNS.LOG_ERROR)
            self.ready = False
        
        else: self.identity = identity

        self.destination_hexhash = self.__resolve_destination_alias(self.destination_hexhash)

    def __resolve_destination_alias(self, alias):
        def resolve(alias):
            len_match = len(alias) == RNS.Identity.TRUNCATED_HASHLENGTH//8*2
            try: hash_bytes = bytes.fromhex(alias)
            except: hash_bytes = None
            if len_match and hash_bytes: return alias
            else: return self.destination_aliases[alias] if alias in self.destination_aliases else alias

        resolved = resolve(alias)
        return resolved

    def abort(self, reason=None, code=255):
        if not reason: reason = "Unknown reason"
        print(f"git-remote-rns failed: {reason}", file=sys.stderr)
        if self.link: self.link.teardown()
        sys.exit(code)

    def connect_server(self):
        try: destination_hash = bytes.fromhex(self.destination_hexhash)
        except Exception as e: self.abort(f"Invalid destination hash: {e}")

        RNS.log(f"Requesting path to {RNS.prettyhexrep(destination_hash)}", RNS.LOG_DEBUG)
        sys.stderr.write(f"Requesting path..."); sys.stderr.flush()
        if not RNS.Transport.await_path(destination_hash, timeout=self.path_timeout):
            sys.stderr.write(f"\n"); sys.stderr.flush()
            self.abort(f"Could not resolve path to {RNS.prettyhexrep(destination_hash)}")
        
        else:
            RNS.log(f"Path to {RNS.prettyhexrep(destination_hash)} resolved", RNS.LOG_DEBUG);
            sys.stderr.write(f"\rPath resolved     "); sys.stderr.flush()

        self.remote_identity = RNS.Identity.recall(destination_hash)
        if not self.remote_identity: self.abort("Could not recall remote identity. Is the server announcing?")

        sys.stderr.write(f"\rEstablishing link..."); sys.stderr.flush()
        self.destination = RNS.Destination(self.remote_identity, RNS.Destination.OUT, RNS.Destination.SINGLE, APP_NAME, "repositories")
        self.link = RNS.Link(self.destination)
        self.link.set_link_established_callback(self.link_established)
        self.link.set_link_closed_callback(self.link_closed)

    def link_established(self, link):
        RNS.log(f"Link established, identifying...", RNS.LOG_DEBUG)
        sys.stderr.write(f"\rLink established with remote\n"); sys.stderr.flush()
        link.identify(self.identity)
        self.link_ready = True

    def link_closed(self, link):
        RNS.log(f"Link was closed", RNS.LOG_DEBUG)
        if not self.link_ready: self.link_failed = True

    def _on_progress(self, transfer_instance):
        if hasattr(transfer_instance, "progress"):
            self.response_progress      = transfer_instance.progress
            self.response_size          = transfer_instance.response_size
            self.response_transfer_size = transfer_instance.response_transfer_size
        
        elif hasattr(transfer_instance, "get_progress") and callable(transfer_instance.get_progress):
            self.response_progress      = transfer_instance.get_progress()
            self.response_size          = transfer_instance.total_size
            self.response_transfer_size = transfer_instance.size
        
        now = time.time()
        if self.progress_updated_at == None: self.progress_updated_at = now

        if now > self.progress_updated_at+1:
            td = now - self.progress_updated_at
            pd = self.response_progress - self.previous_progress
            bd = pd*self.response_size if self.response_size else 0
            self.response_speed = (bd/td)*8 if td > 0 else 0
            self.previous_progress = self.response_progress
            self.progress_updated_at = now
            
            # Report progress to git via stderr
            if self.progress_enabled and self.response_size:
                percent = round(self.response_progress * 100, 1)
                size = self.response_size
                rxd = size*self.response_progress
                speed_kbps = (self.response_speed / 1000) if hasattr(self, 'response_speed') else 0
                sys.stderr.write(f"Transferring: {percent}% ({RNS.prettysize(rxd)}/{RNS.prettysize(size)}) {RNS.prettyspeed(self.response_speed)}          \r")
                sys.stderr.flush()

    ################################
    # Synchronous Request Wrappers #
    ################################

    def _response_ready(self, request_receipt):
        self.request_response = request_receipt.response
        self.response_metadata = request_receipt.metadata

        if hasattr(self.request_response, "read") and callable(self.request_response.read):
            response_path = self.request_response.name
            base_name = os.path.basename(response_path)
            retained_path = os.path.join(self.tmp_dir.name, base_name)
            shutil.move(response_path, retained_path)
            self.request_response = open(retained_path, "rb")

        self.request_event.set()

    def _response_failed(self, request_receipt=None):
        self.request_response = None
        self.request_event.set()

    def send_request(self, path, data, timeout=7200):
        if not self.link_ready: self.abort("Link not ready for request")
        
        self.request_event.clear()
        self.request_response  = None
        self.response_metadata = None
        self.previous_progress = 0
        self.progress_updated_at = None
        
        RNS.log(f"Sending request: {path}", RNS.LOG_DEBUG)
        request_receipt = self.link.request(path, data, progress_callback=self._on_progress, response_callback=self._response_ready, failed_callback=self._response_failed, timeout=timeout)
        if request_receipt.resource: request_receipt.resource.progress_callback(self._on_progress)
        self.request_event.wait(timeout=timeout)
        
        if self.request_response is None: self.abort("Request failed or timed out")
        RNS.log(f"Got response for: {path}", RNS.LOG_DEBUG)
        
        return self.request_response, self.response_metadata

    #############################
    # Git Helper Protocol Logic #
    #############################

    def _detach_stdout(self):
        sys.stdout = open(os.devnull, "w")
        sys.stderr = open(os.devnull, "w")

    def run(self):
        try: self.connect_server()
        except Exception as e: self.abort(str(e))

        timeout = self.link_timeout
        while not self.link_ready and not self.link_failed and timeout > 0:
            time.sleep(0.5)
            timeout -= 1

        if not self.link_ready: self.abort("Failed to establish link")

        self.progress_enabled = False

        git_stdin  = sys.stdin
        git_stdout = sys.stdout
        git_stderr = sys.stderr

        fetch_queue = []
        push_queue  = []

        while True:
            line = git_stdin.readline()
            if not line: break

            line = line.strip()
            if line == "capabilities":
                git_stdout.write("list\n")
                git_stdout.write("fetch\n")
                git_stdout.write("push\n")
                git_stdout.write("option\n")
                git_stdout.write("\n")
                git_stdout.flush()

            elif line == "list": self.handle_git_list(git_stdout)

            elif line.startswith("list "): self.handle_git_list(git_stdout, for_push=True) # List for push

            elif line.startswith("option"):
                # Line format: option <name> <value>
                parts = line.split(maxsplit=2)
                opt_name = parts[1] if len(parts) > 1 else ""
                opt_value = parts[2] if len(parts) > 2 else ""
                
                if opt_name == "progress": self.progress_enabled = opt_value.lower() in ("true", "1", "yes"); git_stdout.write("ok\n")
                else: git_stdout.write("unsupported\n")
                
                git_stdout.flush()

            elif line.startswith("fetch"):
                # Line format: fetch <sha> <ref>
                parts = line.split()
                sha = parts[1]
                ref = parts[2]
                # Avoid duplicates in the same batch - TODO: Re-evaluate this
                if (sha, ref) not in fetch_queue: fetch_queue.append((sha, ref))
                push_queue = []

            elif line.startswith("push"):
                # Line format: push <local_ref>:<remote_ref>
                parts = line.split()
                refspec = parts[1]
                local_ref, remote_ref = refspec.split(":", 1)
                push_queue.append((local_ref, remote_ref))
                fetch_queue = []

            elif line == "": # End of batch
                try:
                    self.process_fetch_queue(fetch_queue, git_stdout, self.progress_enabled, self.ref_batch_size)
                    self.process_push_queue(push_queue, git_stdout, git_stderr, self.progress_enabled)
                    fetch_queue = []
                    push_queue = []
                    git_stdout.write("\n")
                    git_stdout.flush()
                
                except BrokenPipeError:
                    self._detach_stdout()
                    RNS.log("Git closed connection, exiting", RNS.LOG_DEBUG)
                    break

            else: self.abort(f"Unknown Git command: {line}")

        try: sys.stdout.flush()
        except BrokenPipeError: pass

        if self.link: self.link.teardown()

    def handle_git_list(self, git_stdout, for_push=False):
        RNS.log("Handle git list" + (" for-push" if for_push else ""), RNS.LOG_DEBUG)
        request_data = {self.IDX_REPOSITORY: self.repo_path, "for_push": for_push}
        response, metadata = self.send_request(self.PATH_LIST, request_data)

        if not response or not isinstance(response, bytes): self.abort("Invalid list response from server")

        status_byte = response[0]
        payload = response[1:]

        if status_byte != 0: self.abort(f"Server refused list: {payload.decode('utf-8', errors='ignore')}")

        response_text = payload.decode("utf-8")

        self.remote_refs = {}
        for line in response_text.split("\n"):
            line = line.strip()
            if not line: continue
            parts = line.split(" ", 1)
            if len(parts) == 2:
                sha, ref_name = parts
                if ref_name == "HEAD": continue
                self.remote_refs[ref_name] = sha

        git_stdout.write(response_text)
        git_stdout.write("\n") # Required to terminate list
        git_stdout.flush()

    def escape_for_stdout(self, value):
        if isinstance(value, bytes): value = value.decode('utf-8', errors='replace')

        escaped = '"'
        for char in value:
            if   char == '\\': escaped += '\\\\'
            elif char == '"': escaped += '\\"'
            elif char == '\n': escaped += '\\n'
            elif char == '\t': escaped += '\\t'
            elif char == '\r': escaped += '\\r'
            elif ord(char) < 32 or ord(char) > 126: escaped += f'\\x{ord(char):02x}'
            else: escaped += char
        
        return escaped + '"'

    def process_fetch_queue(self, fetch_queue, git_stdout, progress_enabled=False, ref_batch_size=REF_BATCH_SIZE):
        import tempfile
        import subprocess

        if not fetch_queue: return

        # Build a global have list from all remote refs that the client already has objects for
        have_shas = []
        for sha in self.remote_refs.values():
            try:
                result = subprocess.run(["git", "cat-file", "-t", sha], capture_output=True, check=False)
                if result.returncode == 0: have_shas.append(sha)
            
            except Exception as e: RNS.log(f"Could not verify remote SHA {sha} locally: {e}", RNS.LOG_WARNING)

        while fetch_queue:
            batch = fetch_queue[:ref_batch_size]
            fetch_queue = fetch_queue[ref_batch_size:]

            refs_list = []
            for sha, ref in batch:
                ref_entry = {"sha": sha, "ref": ref}
                try:
                    # Attempt to get local ref SHA for incremental bundle generation on remote
                    result = subprocess.run(["git", "rev-parse", ref], capture_output=True, text=True, check=False)
                    if result.returncode == 0:
                        local_sha = result.stdout.strip()
                        if local_sha != sha: ref_entry["have"] = local_sha

                except Exception as e:
                    RNS.log(f"Could not resolve local SHA for {ref} during fetch enumeration, getting full history for this ref: {e}", RNS.LOG_WARNING)

                refs_list.append(ref_entry)

            ref_names = [ref for _, ref in batch]
            RNS.log(f"Fetching batch of {len(refs_list)} refs: {ref_names} (have {len(have_shas)} common objects)", RNS.LOG_DEBUG)

            request_data = { self.IDX_REPOSITORY: self.repo_path, "refs": refs_list }
            if have_shas: request_data["have"] = have_shas

            response, metadata = self.send_request(self.PATH_FETCH, request_data)

            if not response: self.abort(f"No data in fetch response for batch")
            if not metadata:
                if not isinstance(response, bytes): self.abort(f"Invalid fetch response for batch")
                status_byte = response[0]

                if status_byte == 0:
                    RNS.log(f"Server returned empty bundle, all objects already exist locally", RNS.LOG_DEBUG)
                    continue

                else:
                    error_msg = response[1:].decode('utf-8', errors='ignore')
                    self.abort(f"Fetch failed for batch: {error_msg}")

            else:
                if not self.IDX_RESULT_CODE in metadata: self.abort(f"No result metadata on bundle response")
                status_byte = metadata[self.IDX_RESULT_CODE]
                if status_byte == 0: bundle_path = response.name
                else: self.abort(f"Unknown remote state for batch ref fetch")

                if progress_enabled:
                    size = os.stat(bundle_path).st_size
                    sys.stderr.write(f"Transferring: 100% ({RNS.prettysize(size)}).                       \n")
                    sys.stderr.flush()

                stderr_arg = sys.stderr if progress_enabled else subprocess.DEVNULL

                verify_cmd = ["git", "bundle", "verify", "-q", bundle_path]
                verify_result = subprocess.run(verify_cmd, stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)

                if verify_result.returncode != 0: self.abort(f"Bundle verification failed for batch")

                unbundle_cmd = ["git", "bundle", "unbundle"]
                if progress_enabled: unbundle_cmd.append("--progress")
                unbundle_cmd.append(bundle_path)

                unbundle_result = subprocess.run(unbundle_cmd, stderr=stderr_arg, stdout=subprocess.DEVNULL)

                if unbundle_result.returncode != 0: self.abort(f"Bundle unbundle failed for batch: Non-zero return code")

    def process_push_queue(self, push_queue, git_stdout, git_stderr, progress_enabled=False):
        import tempfile
        import subprocess

        for local_ref, remote_ref in push_queue:
            RNS.log(f"Pushing {local_ref} to {remote_ref}", RNS.LOG_DEBUG)

            # Handle potential deletions
            if not local_ref or local_ref == "":
                request_data = { self.IDX_REPOSITORY: self.repo_path, "ref": remote_ref }
                response, metadata = self.send_request(self.PATH_DELETE, request_data)

                if not response or not isinstance(response, bytes):
                    git_stdout.write(f"error {remote_ref} {self.escape_for_stdout('No response from server')}\n")
                    git_stdout.flush()
                    continue

                status_byte = response[0]
                if status_byte != 0:
                    error_msg = response[1:].decode("utf-8", errors="ignore")
                    git_stdout.write(f"error {remote_ref} {self.escape_for_stdout(error_msg)}\n")
                    git_stdout.flush()
                    continue

                git_stdout.write(f"ok {remote_ref}\n")
                git_stdout.flush()
                continue

            force = local_ref.startswith("+")
            if force: local_ref = local_ref[1:]

            stderr_arg = sys.stderr if progress_enabled else subprocess.DEVNULL

            # Resolve the SHA that local_ref points to
            sha_result = subprocess.run(["git", "rev-parse", local_ref], capture_output=True, text=True, check=False)
            if sha_result.returncode != 0:
                error_msg = f"Could not resolve local ref {local_ref}"
                git_stdout.write(f"error {remote_ref} {self.escape_for_stdout(error_msg)}\n")
                git_stdout.flush()
                continue

            local_sha = sha_result.stdout.strip()

            bundle_empty = False
            with tempfile.TemporaryDirectory() as tmpdir:
                bundle_path = tmpdir + "/push.bundle"

                create_cmd = ["git", "bundle", "create", bundle_path, local_ref]

                # Exclude all remote ref SHAs that exist locally, so the
                # bundle only contains objects the remote doesn't already have
                exclude_count = 0
                for sha in self.remote_refs.values():
                    try:
                        # We need to verify each SHA actually exists locally, since git
                        # bundle create will fail if a ^<sha> argument references an object
                        # not present in the local repository.
                        result = subprocess.run(["git", "cat-file", "-t", sha], capture_output=True, check=False)
                        if result.returncode == 0:
                            create_cmd.append(f"^{sha}")
                            exclude_count += 1
                    
                    except Exception as e: RNS.log(f"Could not verify remote SHA {sha} locally: {e}", RNS.LOG_WARNING)

                RNS.log(f"Excluding {exclude_count}/{len(self.remote_refs)} remote refs for {local_ref}", RNS.LOG_DEBUG)

                if progress_enabled: create_cmd.insert(3, "--progress")

                create_result = subprocess.run(create_cmd, capture_output=True, text=True, check=False)

                if create_result.returncode == 0:
                    if create_result.stderr:
                        # git_stderr.write(create_result.stderr)
                        pass
                else:
                    if "empty bundle" in create_result.stderr.lower():
                        # All objects reachable from local_ref already exist on
                        # the remote. In this case, no bundle is needed and we can
                        # update the ref directly via the operations path instead.
                        bundle_empty = True
                        RNS.log(f"Empty bundle for {local_ref}, all objects already on remote", RNS.LOG_DEBUG)

                    else:
                        if progress_enabled and create_result.stderr: git_stderr.write(create_result.stderr)
                        error_msg = "Bundle creation failed"
                        git_stdout.write(f"error {remote_ref} {self.escape_for_stdout(error_msg)}\n")
                        git_stdout.flush()
                        continue

                if not bundle_empty:
                    with open(bundle_path, "rb") as f: bundle_data = f.read()

                    request_data = { self.IDX_REPOSITORY: self.repo_path, "local_ref": local_ref, "remote_ref": remote_ref,
                                     "force": force, "bundle": bundle_data }
                    
                    response, metadata = self.send_request(self.PATH_PUSH, request_data)

                    if not response or not isinstance(response, bytes):
                        git_stdout.write(f"error {remote_ref} {self.escape_for_stdout('No response from server')}\n")
                        git_stdout.flush()
                        continue

                    status_byte = response[0]
                    if status_byte != 0:
                        error_msg = response[1:].decode('utf-8', errors='ignore')
                        git_stdout.write(f"error {remote_ref} {self.escape_for_stdout(error_msg)}\n")
                        git_stdout.flush()
                        continue

            # When all reachable objects already exist on the remote, send a
            # direct ref update operation instead of a bundle.
            if bundle_empty:
                operation    = {"action": "update_ref", "ref": remote_ref, "sha": local_sha, "force": force}
                request_data = { self.IDX_REPOSITORY: self.repo_path,
                                 "operations": [operation] }

                response, metadata = self.send_request(self.PATH_PUSH, request_data)

                if not response or not isinstance(response, bytes):
                    git_stdout.write(f"error {remote_ref} {self.escape_for_stdout('No response from server')}\n")
                    git_stdout.flush()
                    continue

                status_byte = response[0]
                if status_byte != 0:
                    error_msg = response[1:].decode('utf-8', errors='ignore')
                    git_stdout.write(f"error {remote_ref} {self.escape_for_stdout(error_msg)}\n")
                    git_stdout.flush()
                    continue

            git_stdout.write(f"ok {remote_ref}\n")
            git_stdout.flush()


__default_rngit_config__ = '''# This is the default rngit client config file.

[client]

# You can control the batch size of ref transfers
# using the ref_batch_size directive:

ref_batch_size = 25


[aliases]

# You can define aliases for commonly used destination
# hashes in this section. Each line must be in the format
# aliased_name = DESTINATION_HASH
#
# These hashes are used for resolving remote destinations.
# For rngit node permissions and identity resolution,
# aliases must be defined in ~/.rngit/config.

# my_node = 063d38912bffc850af4a1b8a270a9d85
# bobs_node = 714981d03e41deda0e4468cb274414cc


[logging]
# Valid log levels are 0 through 7:
#   0: Log only critical information
#   1: Log errors and lower log levels
#   2: Log warnings and lower log levels
#   3: Log notices and lower log levels
#   4: Log info and lower (this is the default)
#   5: Verbose logging
#   6: Debug logging
#   7: Extreme logging

loglevel = 4

'''.splitlines()

if __name__ == "__main__": main()